<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2022 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2022.1/2316806
 */
namespace Reviews\Listener;

use Application\Config\IConfigDefinition;
use Application\Log\SwarmLogger;
use Events\Listener\AbstractEventListener;
use Laminas\EventManager\Event;
use Application\Config\ConfigManager;
use Application\Config\ConfigException;
use Reviews\Model\IReview;
use Reviews\Service\IStatistics;

/**
 * Class Statistics. A listener to calculate review statistics
 * @package Reviews\Listener
 */
class Statistics extends AbstractEventListener
{
    /**
     * Calculate a complexity value on review content change
     * @param Event $event  the event
     */
    public function reviewChanged(Event $event)
    {
        $data = $event->getParam('data');
        // task.review is fired for state change etc., we only want to recalculate if an update has occurred
        // or a file or files have been deleted from the change list
        if ($data && (isset($data['updateFromChange']) || isset($data[IReview::DELFROMCHANGE]))) {
            $class  = get_class($this);
            $logger = $this->services->get(SwarmLogger::SERVICE);
            // We want to deal with the review that the review listener has been dealing with as we are planning to
            // make changes and save
            $review = $event->getParam('review');
            $logger->trace(sprintf("%s: Calculating complexity for review [%s]", $class, $review->getId()));
            $complexityService = $this->services->get(IStatistics::COMPLEXITY_SERVICE);
            $complexityService->calculateComplexity($review);
            $logger->trace(sprintf("%s: Calculated complexity for review [%s]", $class, $review->getId()));
        }
    }

    /**
     * Attaches this event only if the complexity calculation configuration value is set to default to use the
     * Swarm implementation
     * @param mixed $eventName      the event name
     * @param array $eventDetail    the event detail
     * @return bool true if the complexity calculation configuration value is 'default', otherwise false
     * @throws ConfigException
     */
    public function shouldAttach($eventName, $eventDetail)
    {
        $config = $this->services->get(IConfigDefinition::CONFIG);
        return ConfigManager::getValue(
            $config,
            IConfigDefinition::REVIEWS_COMPLEXITY_CALCULATION,
            IConfigDefinition::DEFAULT
        ) === IConfigDefinition::DEFAULT;
    }
}
